@@ -15,6 +15,7 @@ struct Device
   int secLength;
   int tracks;
   int sectrk;
+  off_t offset;
 #if HAVE_LIBDSK_H
   DSK_PDRIVER   dev;
   DSK_GEOMETRY geom; 
@@ -27,7 +28,7 @@ struct Device
 };
 
 const char *Device_open(struct Device *self, const char *filename, int mode, const char *deviceOpts);
-void Device_setGeometry(struct Device *self, int secLength, int sectrk, int tracks);
+void Device_setGeometry(struct Device *self, int secLength, int sectrk, int tracks, off_t offset);
 const char *Device_close(struct Device *self);
 const char *Device_readSector(const struct Device *self, int track, int sector, char *buf);
 const char *Device_writeSector(const struct Device *self, int track, int sector, const char *buf);
